import 'dart:convert';
import 'dart:io';
import 'dart:core';
import 'package:simplewebrtc_flutter_webrtc_shim/webrtc.dart';
import 'package:http/http.dart';
import "package:json_rpc_2/json_rpc_2.dart";
import "package:web_socket_channel/io.dart";

typedef void ConnectionStateCallback(dynamic event);
typedef void RoomStateCallback(dynamic event);
typedef void PeerUpdateCallback(dynamic event);
typedef void MediaStreamCallback(MediaStream stream);

class SimpleWebRTCSDK {
  static final String sdkVersion = "0.0.12-flutter";
  final String configUrl;
  Map<String, dynamic> config;
  Peer rpc;
  JsonEncoder encoder = new JsonEncoder();
  JsonDecoder decoder = new JsonDecoder();
  MediaStream localStream;
  List<MediaStream> remoteStreams = [];
  Map<String, String> rooms = {};
  Map<String, String> roomAddresses = {};
  Map<String, dynamic> peers = {};
  Map<String, RTCPeerConnection> peerConnections = {};

  String proxyUrl;
  IOWebSocketChannel _proxySocket;

  bool _intendMuteRemote = false;
  bool _currentlySpeaking = false;

  ConnectionStateCallback onConnectionState;
  RoomStateCallback onRoomState;
  PeerUpdateCallback onPeerState;
  MediaStreamCallback onRemoteMediaAdded;
  MediaStreamCallback onRemoteMediaRemoved;
  MediaStreamCallback onLocalMediaAdded;
  MediaStreamCallback onLocalMediaRemoved;

  SimpleWebRTCSDK({this.configUrl, this.proxyUrl});

  bool get isMuted {
    if (this.localStream == null) {
      return true;
    }
    List<MediaStreamTrack> audio = this.localStream.getAudioTracks();
    if (audio[0] == null) {
      return true;
    }
    return !audio[0].enabled;
  }

  bool get remoteIsMuted {
    return this._intendMuteRemote;
  }

  Future<void> getUserMedia({bool audio, bool video}) async {
    this.localStream =
        await navigator.getUserMedia({'audio': audio, 'video': video});

    await this.localStream.getMediaTracks();
    this.localStream.getAudioTracks()[0].enableSpeakerphone(true);

    this._monitorAudioLevel(
        threshold: 1000, pollingInterval: 500, previousAudioInput: 0);

    if (this.onLocalMediaAdded != null) {
      this.onLocalMediaAdded(this.localStream);
    }
    return this.localStream;
  }

  Future<void> endUserMedia() async {
    if (this.localStream != null) {
      if (this.onLocalMediaRemoved != null) {
        this.onLocalMediaRemoved(this.localStream);
      }
      await this.localStream.dispose();
      this.localStream = null;
    }
  }

  Future<void> connect() async {
    Response resp;
    try {
      String body = this.encoder.convert({'clientVersion': SimpleWebRTCSDK.sdkVersion});
      resp = await post(this.configUrl, body: body);
    } catch (err) {
      if (this.onConnectionState != null) {
        this.onConnectionState({"connectionState": "failed"});
      }
      print("Failed to retrieve SimpleWebRTC client configuration");
      throw err;
    }

    this.config = this.decoder.convert(resp.body);
    print("Received SimpleWebRTC client configuration: ${this.config}");
    this.proxyUrl = this.config["proxyUrl"] ?? this.proxyUrl;

    this._proxySocket = IOWebSocketChannel.connect(this.proxyUrl);
    this.rpc = Peer(this._proxySocket.cast<String>());
    this._setupRPCHandlers();

    this.rpc.listen().then((_) {
      print('Closed SimpleWebRTC connection');
    }).catchError((err) {
      this.endUserMedia();
      if (this.onConnectionState != null) {
        this.onConnectionState({"connectionState": "disconnected"});
      }
      print('SimpleWebRTC connection closed from an error');
      print(err);
    });

    return this.rpc.sendRequest('CONNECT', [
      {
        "jid": this.config['userId'],
        "password": this.config['credential'],
        "resource": this.config['id'],
        "url": this.config['signalingUrl']
      }
    ]).then((res) {
      if (this.onConnectionState != null) {
        this.onConnectionState({"connectionState": "connected"});
      }
      return res;
    }).catchError((err) {
      if (this.onConnectionState != null) {
        this.onConnectionState({"connectionState": "failed"});
      }
      print("Failed to connect to SimpleWebRTC proxy service");
      throw err;
    });
  }

  Future<void> joinRoom(String roomName) async {
    String body = this.encoder.convert({'name': roomName});
    Response resp;
    try {
      print("Fetching SimpleWebRTC room configuration for room: $roomName");
      resp = await post(this.config['roomConfigUrl'], body: body, headers: {
        HttpHeaders.authorizationHeader: "Bearer ${this.config['credential']}",
        HttpHeaders.contentTypeHeader: 'application/json'
      });
    } catch (err) {
      print("Failed to retrieve SimpleWebRTC room configuration");
      throw err;
    }

    var roomConfig = this.decoder.convert(resp.body);
    print(
        "Received SimpleWebRTC configuration for room $roomName: $roomConfig");

    this.rooms[roomName] = roomConfig['roomAddress'];

    print(
        "Joining SimpleWebRTC room $roomName at ${roomConfig['roomAddress']}");
    return this.rpc.sendRequest('JOIN_ROOM', [
      {"roomAddress": roomConfig['roomAddress']}
    ]).then((res) {
      if (this.onRoomState != null) {
        this.onRoomState({
          "roomName": roomName,
          "roomAddress": roomConfig['roomAddress'],
          "joined": true
        });
      }
      return res;
    });
  }

  Future<void> leaveRoom(String roomName) async {
    // This is assuming a single joined room at a time
    this.peerConnections.forEach((sessionId, pc) {
      pc.close();
    });
    this.peerConnections = {};

    return this.rpc.sendRequest('LEAVE_ROOM', [
      {"roomAddress": this.rooms[roomName]}
    ]);
  }

  Future<void> mute() async {
    if (this.localStream != null) {
      this.localStream.getAudioTracks()[0].enabled = false;

      if (this._currentlySpeaking) {
        this._sendSpeakingNotification(false);
      }
      this._currentlySpeaking = false;

      this.peerConnections.forEach((sessionId, pc) {
        if (pc.getLocalStreams().length > 0) {
          this.rpc.sendNotification('MUTE', [
            {"sessionId": sessionId}
          ]);
        }
      });
    }
  }

  Future<void> unmute() async {
    if (this.localStream != null) {
      this.localStream.getAudioTracks()[0].enabled = true;
      this.peerConnections.forEach((sessionId, pc) {
        if (pc.getLocalStreams().length > 0) {
          this.rpc.sendNotification('UNMUTE', [
            {"sessionId": sessionId}
          ]);
        }
      });
    }
  }

  Future<void> muteRemote() async {
    this._intendMuteRemote = true;
    this.peerConnections.forEach((sessionId, pc) async {
      List<MediaStream> streams = pc.getRemoteStreams();
      if (streams.length > 0 && streams[0] != null) {
        await streams[0].getMediaTracks();
        List<MediaStreamTrack> tracks = streams[0].getAudioTracks();
        tracks.forEach((track) {
          track.enabled = false;
        });
      }
    });
  }

  Future<void> unmuteRemote() async {
    this._intendMuteRemote = false;
    this.peerConnections.forEach((sessionId, pc) async {
      List<MediaStream> streams = pc.getRemoteStreams();
      if (streams.length > 0 && streams[0] != null) {
        await streams[0].getMediaTracks();
        List<MediaStreamTrack> tracks = streams[0].getAudioTracks();
        tracks.forEach((track) {
          track.enabled = true;
        });
      }
    });
  }

  Future<void> disconnect() async {
    this.endUserMedia();
    this.rpc.close();
    if (this.onConnectionState != null) {
      this.onConnectionState({"connectionState": "disconnected"});
    }
  }

  void _setupRPCHandlers() {
    this.rpc.registerMethod("SELF_STATUS", (params) {
      print("SELF_STATUS ${params.asMap}");
    });

    this.rpc.registerMethod("PEER_STATUS", (params) {
      print("PEER_STATUS ${params.asMap}");

      String peerAddress = params.asMap['peerAddress'];

      var existing = this.peers[peerAddress];
      this._updatePeer(peerAddress, params.asMap);

      if (existing == null) {
        print("Initiating outgoing session to $peerAddress");
        this.rpc.sendNotification('SESSION_INITIATE', [
          {"peerAddress": peerAddress}
        ]);
      }
    });

    this.rpc.registerMethod("PEER_OFFLINE", (params) {
      print("PEER_OFFLINE ${params.asMap}");

      this.peers.remove(params.asMap['peerAddress']);

      if (this.onPeerState != null) {
        this.onPeerState({
          "online": false,
          "speaking": false,
          "remoteMuted": false,
          "peerAddress": params.asMap['peerAddress']
        });
      }
    });

    this.rpc.registerMethod("LEFT_ROOM", (params) {
      print("LEFT_ROOM ${params.asMap}");

      String roomName = this.roomAddresses.remove(params.asMap['roomAddress']);
      if (roomName != null) {
        this.rooms.remove(roomName);
      }

      this.peers.removeWhere((peerAddress, _) {
        return peerAddress.startsWith("${params.asMap['roomAddress']}/");
      });

      // This is assuming a single joined room at a time
      this.peerConnections.forEach((sessionId, pc) {
        if (this.onRemoteMediaRemoved != null) {
          List<MediaStream> streams = pc.getRemoteStreams();
          streams.forEach((stream) {
            this.onRemoteMediaRemoved(stream);
          });
        }
        pc.close();
      });
      this.peerConnections = {};

      if (this.onRoomState != null) {
        this.onRoomState({
          "roomName": roomName,
          "roomAddress": params.asMap['roomAddress'],
          "joined": false
        });
      }
    });

    this.rpc.registerMethod("CREATE_OFFER", (params) async {
      print("CREATE_OFFER ${params.asMap}");
      String sessionId = params.asMap['sessionId'];

      if (this.localStream == null) {
        await this.getUserMedia(audio: true);
      }

      RTCPeerConnection pc = await this._createPeerConnection(sessionId);
      this.peerConnections[sessionId] = pc;

      await pc.addStream(this.localStream);

      RTCSessionDescription offer = await pc.createOffer(
          {"offerToReceiveAudio": true, "offerToReceiveVideo": false});

      return {"type": offer.type, "sdp": offer.sdp};
    });

    this.rpc.registerMethod("CREATE_ANSWER", (params) async {
      print("CREATE_ANSWER ${params.asMap}");

      String sessionId = params.asMap['sessionId'];
      RTCPeerConnection pc = this.peerConnections[sessionId];

      if (pc == null) {
        throw new Exception("Missing peer connection");
      }

      RTCSessionDescription answer = await pc.createAnswer(
          {'offerToReceiveAudio': true, 'offerToReceiveVideo': false});

      return {'sdp': answer.sdp, 'type': answer.type};
    });

    this.rpc.registerMethod("SET_LOCAL_DESCRIPTION", (params) async {
      print("SET_LOCAL_DESCRIPTION ${params.asMap}");

      String sessionId = params.asMap['sessionId'];
      RTCPeerConnection pc = this.peerConnections[sessionId];

      if (pc == null) {
        throw new Exception("Missing peer connection");
      }

      await pc.setLocalDescription(
          RTCSessionDescription(params.asMap['sdp'], params.asMap['type']));
    });

    this.rpc.registerMethod("SET_REMOTE_DESCRIPTION", (params) async {
      print("SET_REMOTE_DESCRIPTION ${params.asMap}");

      String sessionId = params.asMap['sessionId'];
      RTCPeerConnection pc = this.peerConnections[sessionId];

      if (pc == null) {
        pc = await this._createPeerConnection(sessionId);
        this.peerConnections[sessionId] = pc;

        pc.onAddStream = (stream) {
          if (this._intendMuteRemote) {
            this.muteRemote();
          }
          if (this.onRemoteMediaAdded != null) {
            this.onRemoteMediaAdded(stream);
          }
        };
      } else {
        Future.delayed(const Duration(milliseconds: 1000), () {
          List<MediaStreamTrack> tracks = this.localStream.getAudioTracks();
          if (!tracks[0].enabled) {
            this.rpc.sendNotification('MUTE', [
              {"sessionId": sessionId}
            ]);
          }
        });
      }

      await pc.setRemoteDescription(
          RTCSessionDescription(params.asMap['sdp'], params.asMap['type']));
    });

    this.rpc.registerMethod("ADD_ICE_CANDIDATE", (params) async {
      print("ADD_ICE_CANDIDATE ${params.asMap}");

      String sessionId = params.asMap["sessionId"];
      RTCPeerConnection pc = this.peerConnections[sessionId];

      if (pc == null) {
        return;
      }

      return pc.addCandidate(new RTCIceCandidate(
        params.asMap['candidate'],
        params.asMap['sdpMid'],
        0,
      ));
    });

    this.rpc.registerMethod("SESSION_INITIATE", (params) async {
      print("SESSION_INITIATE ${params.asMap}");

      String sessionId = params.asMap["sessionId"];

      Future.delayed(const Duration(milliseconds: 1000), () {
        this.rpc.sendNotification('SESSION_ACCEPT', [
          {'sessionId': sessionId}
        ]);
      });
    });

    this.rpc.registerMethod("SESSION_TERMINATE", (params) {
      print("SESSION_TERMINATE ${params.asMap}");

      String sessionId = params.asMap["sessionId"];
      if (this.peerConnections[sessionId] != null) {
        RTCPeerConnection pc = this.peerConnections[sessionId];
        if (this.onRemoteMediaRemoved != null) {
          List<MediaStream> streams = pc.getRemoteStreams();
          streams.forEach((stream) {
            this.onRemoteMediaRemoved(stream);
          });
        }
        pc.close();
        this.peerConnections.remove(sessionId);
      }
    });

    this.rpc.registerMethod("MUTE", (params) {
      print("MUTE ${params.asMap}");

      String peerAddress = params.asMap['peerAddress'];
      this._updatePeer(peerAddress, {"remoteMuted": true});
    });

    this.rpc.registerMethod("UNMUTE", (params) {
      print("UNMUTE ${params.asMap}");

      String peerAddress = params.asMap['peerAddress'];
      this._updatePeer(peerAddress, {"remoteMuted": false});
    });

    this.rpc.registerMethod("SPEAKING_STATUS", (params) {
      print("SPEAKING_STATUS ${params.asMap}");

      String peerAddress = params.asMap['peerAddress'];
      this._updatePeer(peerAddress, {"speaking": params.asMap['speaking']});
    });
  }

  Future<RTCPeerConnection> _createPeerConnection(String sessionId) async {
    List<dynamic> iceServerConfigs = this.config['iceServers'];
    List<dynamic> iceServers = [];

    iceServerConfigs.forEach((server) {
      String host = server['host'];
      if (host.contains(':')) {
        host = "[$host]";
      }
      String uri = "${server['type']}:$host";
      if (server['port'] != null) {
        uri += ":${server['port']}";
      }
      if (server['transport'] != null) {
        uri += "?transport=${server['transport']}";
      }

      var ice = {};
      ice['url'] = uri;
      if (server['username'] != null) {
        ice['username'] = server['username'];
      }
      if (server['password'] != null) {
        ice['credential'] = server['password'];
      }

      iceServers.add(ice);
    });

    RTCPeerConnection pc = await createPeerConnection(
        {"iceServers": iceServers, "sdpSemantics": 'plan-b'}, {});

    pc.onIceCandidate = (candidate) {
      if (candidate == null) {
        return;
      }

      print("Sending ICE_CANDIDATE $candidate");
      this.rpc.sendNotification('ICE_CANDIDATE', [
        {
          "sdpMid": candidate.sdpMid,
          "candidate": candidate.candidate,
          "sessionId": sessionId
        }
      ]);
    };

    return pc;
  }

  Future<void> _monitorAudioLevel(
      {num threshold, int pollingInterval, int previousAudioInput}) async {
    RTCPeerConnection monitor;
    this.peerConnections.forEach((sessionId, pc) {
      if (pc.getLocalStreams().length > 0 &&
          pc.signalingState != RTCSignalingState.RTCSignalingStateClosed) {
        monitor = pc;
      }
    });

    int audioInputLevel = 0;

    if (monitor != null &&
        this.localStream != null &&
        this.localStream.getAudioTracks()[0].enabled) {
      try {
        List stats = await monitor.getStats();
        stats.forEach((stat) {
          if (stat.type == 'ssrc') {
            audioInputLevel =
                int.parse(stat.values['audioInputLevel'], radix: 10);
          }
        });

        if (audioInputLevel >= threshold && previousAudioInput < threshold) {
          this._currentlySpeaking = true;
          this._sendSpeakingNotification(true);

        }
        if (audioInputLevel < threshold && previousAudioInput >= threshold) {
          this._currentlySpeaking = false;
          this._sendSpeakingNotification(false);
        }
      } catch (err) {
        print(err);
      }
    }

    if (this.localStream != null) {
      Future.delayed(Duration(milliseconds: pollingInterval), () {
        this._monitorAudioLevel(
            threshold: threshold,
            pollingInterval: pollingInterval,
            previousAudioInput: audioInputLevel);
      });
    }
  }

  void _updatePeer(String peerAddress, Map<String, dynamic> data) {
    if (this.peers[peerAddress] == null) {
      this.peers[peerAddress] = {};
    }
    data.forEach((key, value) {
      this.peers[peerAddress][key] = value;
    });
    if (this.onPeerState != null) {
      this.onPeerState({
        "peerAddress": peerAddress,
        "online": true,
        "remoteMuted": this.peers[peerAddress]["remoteMuted"] ?? false,
        "speaking": this.peers[peerAddress]["speaking"] ?? false
      });
    }
  }

  void _sendSpeakingNotification(bool speaking) {
    this.rooms.forEach((roomName, roomAddress) {
      this.rpc.sendNotification('SPEAKING_STATUS', [
        {"roomAddress": roomAddress, "speaking": speaking}
      ]);
    });
  }
}
